/**
 * @file  tds_cls.hpp
 * @brief TDS sampler (K5/VSSP) data handling abstract class
 *
 * @author M.Sekido (NICT Kashima.)
 * @date $Date: 2010/01/02 03:40:37 $
 * @version $Id: tds_cls.hpp,v 1.2 2010/01/02 03:40:37 sekido Exp $
 * $Log: tds_cls.hpp,v $
 * Revision 1.2  2010/01/02 03:40:37  sekido
 * Replace FirstBranch to main
 *
 * Revision 1.1.2.7  2009/11/16 00:20:24  sekido
 * update on Nov15.
 *
 * Revision 1.1.2.6  2009/10/08 14:22:18  sekido
 * DOXY comment debugging finished
 *
 * Revision 1.1.2.5  2009/09/28 10:32:07  sekido
 * sudp_recv2 compile.
 *
 * Revision 1.1.2.4  2009/09/11 10:03:07  sekido
 *  Adjusting sudp_send...
 *
 * Revision 1.1.2.3  2009/09/10 12:55:28  sekido
 * Adjusting for sending out  VDIF/UDP packet.
 *
 * Revision 1.1.2.2  2009/09/09 11:54:33  vlbi
 * sudp_send runnig version is ready.
 *
 * Revision 1.1.2.1  2009/09/09 04:31:15  vlbi
 * sudp_send created and compiled.
 *
 * Revision 1.3  2009/06/02 01:56:54  vlbi
 * Bug fixing for rt-mpcal-dtctr. -O3 option for Makefile.
 *
 * Revision 1.2  2009/06/01 12:44:50  vlbi
 * Corrected error in rt-mpcal-dtctr, which make error in dmesg and break the VSSP32 functionality.
 * Thain workaround was immediate start/stop in  TDS_SMPLR::Read() in tds_sampler.hpp and added
 *  inst.Set_cntrl(0);
 *  smplr.Start_sampling(); and  smplr.Stop_sampling(); in Sokutei() of rt-mpcal-dtctr.cpp.
 *
 * Revision 1.1  2009/05/28 12:19:11  sekido
 * Registering appropriate selected source files.
 *
 * Revision 1.6  2007/09/14 01:24:12  sekido
 * Recovering conflict in tds_cls.hpp
 *
 * Revision 1.5  2007/09/13 06:37:27  sekido
 * On the way to make correlator/ rt-mpcal-dtctr was updated by debug.
 *
 * Revision 1.3  2007/05/14 05:59:59  sekido
 *  -- Append mode added for file output
 *  -- Doxygen comment added
 *  -- cvs keyword added
 *
 * 
 * Copyright (C) NICT Kashima. All rights reserved.
 */


#ifndef __TDS_CLS__
#define __TDS_CLS__

#include <vector>
#include <iostream>
using namespace std;
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <cstring>

extern "C" {
#include <inttypes.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
}

#include "time_lib.hpp"
#include "Logsystem.hpp"

//namespace tds{
#ifdef _VSSP_
const int hbyte = 8; ///<  Old headder size
#else
  const int hbyte = 32; ///< New headder size
#endif

const int FILE_END  =0; ///< FILE_END
const int HD_DETECT_SUCCESS = 1; ///< Define hd-detect success
const int NORMAL =1;     ///< Define normal 
const int HD_DETECT_FAILURE = -1; ///< Define hd-detect failure
const uint32_t MASK_1BIT = 0x01 ; ///< 1bit mask
const uint32_t MASK_2BIT = 0x03 ; ///< 2bit mask
const uint32_t MASK_4BIT = 0x0f ; ///< 4bit mask
const double MSEC = 1000.;  ///< factor of msec
const double KHZ = 1000.;   ///< factor of kHz
const double EXTRA_T = 40.; ///< Extra data for Y (msec)

//  static uint8_t sync_hd[4] = {0xff,0xff,0xff,0xff};
  //  int8_t verbose_level = 0;
//}
//extern uint8_t sync_hd[];

/**
 * A class for handling TDS_data 
 */

class TDS_DAT{
protected:
  uint32_t sfrq;  ///<  sampling rate  (Hz) 
  uint8_t  sbit;  ///< sampling quantization bit 
  uint8_t  nch;   ///< 1ch, or 4ch */
  vector <float> value_tbl;    ///< Value table 
  //  vector <float> farr[4]; /// pointer for data array
  vector <uint8_t>::iterator   bp;   ///<  Buffer position remainder 
  vector <uint8_t>::iterator   bpe;  ///< Buffer end pinter 
  Time_dat t_cur; ///< Current time 
  int time_mode;  ///< 0: time of  headder data, 1: CPU time for t_cur
  /// Interpret (decode) 1 bit data
  const int32_t Interprit_1bit(const uint32_t n_req);
  /// Interpret (decode) 2 bit data
  const int32_t Interprit_2bit(const uint32_t n_req);
  /// Interpret (decode) 4 bit data
  const int32_t Interprit_4bit(const uint32_t n_req);
  /// Interpret (decode) 8 bit data
  const int32_t Interprit_8bit(const uint32_t n_req);
  /// Set Value table 
  void Set_Valtbl(void);
  /// Returns pointer of value table 
  float* pValue_tbl(const int ofst) {return & value_tbl[ofst];}
  /** Abstract function to copy data from *pf to Farr 
   * @param ch: Channel Number
   * @param npt: Number of points
   * @param *pf: pointer of data source
   */
  virtual const int Farr_Set(const int ch, const uint32_t npt, const float *pf)=0;
  /** Abstract function to return test  pfye[ich] + inc -1 < end of farr[ch]
   * @param ich : Channel number 
   * @param inc : Inclement point number */
  virtual bool pFarr_inc(const int ich, const int inc)=0;
  /// Abstract function to return vector iterator of Time array
  virtual vector <double>::iterator& pTarr_tail(void)=0;
  /// Abstract function to returns Elapsed time
  virtual const double Time_elapsed(void)const =0;
  /// Abstract function to retrun the time of Beginning of the file
  virtual const Time_dat & Time_bof(void)const =0;
  //  virtual void Set_Tcur(Time_dat &)=0;
  /// Abstract funciton to add elapsed time
  virtual void Add_time_elapsed(double )=0;
  //  virtual bool pTarr_inc(const int inc)=0;
  /// Returns vector iterator of data begin of the buffer 
  vector <uint8_t>::iterator &Buffer_pointer(void) {return bp;}
  /// Returns vector iterator of data end of the buffer 
  const vector <uint8_t>::iterator &Buffer_End(void) {return bpe;}
 public:
  /// Constructor 
  TDS_DAT(void):sfrq(4000000),sbit(2),nch(1),time_mode(1){
    Set_Valtbl();  t_cur.Set_time_to_now();
    setenv("TZ","GMT",1);tzset();
  }
  virtual ~TDS_DAT(void){;} ///< Destructor 
  /// Abstract function to Open data file 
  virtual const int Open(const char *fnmae)=0;
  /// Abstract function to Read data for a msec of data
  virtual const int32_t Read(const int msec)=0;
  /// Abstract function to return reference of data array
  virtual const vector <float>& Farr(const int ich) const =0;
  /// Abstrct function to return Number of data ponts
  virtual const uint32_t Ndat(void) const =0;
  //  virtual const Time_dat & Time_current(void) const =0;
  /// Returns Quantization bit number by reference 
  const uint8_t & Sbit(void) const {return(sbit);}
  /// Retruns Number of channels by reference 
  const uint8_t & Nch(void) const {return(nch);}
  /// Returns Sampling rate by reference 
  const uint32_t & Sfrq(void) const {return sfrq;}
  /// Set quantization bit number 
  void Set_bit(const int b){sbit=b; Set_Valtbl();}
  /// Set number of channels 
  void Set_nch(const int b){if(b==4) nch=4;else nch=1;}
  /// Set Sampling frequency 
  void Set_sfrq(const int r){
    if(r>1.e4) sfrq=r; else sfrq = r * (r>=40 ? 1000 : 1000000);}
  /// Write information to the file pointer 
  void Info(FILE *fp=stderr);  
  /** Interpretation (decoding) data. 
   *  @param n_req : number of data to be processedb */
  int32_t Interprit_data(const uint32_t n_req);
  /// Returns available data bytes remaining in the buffer.
  const uint32_t Buf_Remain(void) const {return bpe - bp;}
  /// Set Current time
  void Set_Tcur(Time_dat & t){ t_cur = t;}
  /// Returns reference of Current time data 
  const Time_dat & Time_current(void) const { return(t_cur);}

};

inline void TDS_DAT::Info(FILE *fp){
  //  using namespace tds;
  if(sfrq <= 1.e6)
    fprintf(fp, "Obsavation mode: %d kHz %d bit %d ch \n", sfrq/1000000, sbit, nch);
  else 
    fprintf(fp, "Obsavation mode: %d MHz %d bit %d ch \n", sfrq/1000000, sbit, nch);
  //  fprintf(stderr,"Verbose_level= %d\n",Logsystem::Dbg_logger()->Level());
}

inline int32_t TDS_DAT::Interprit_data(const uint32_t n_req){
  if( n_req == 0 ) return 0;
  
  switch( sbit){
  case 1: 
    return(Interprit_1bit(n_req)); break;
  case 2:
    return(Interprit_2bit(n_req)); break;
  case 4:
    return(Interprit_4bit(n_req)); break;
  case 8:
    return(Interprit_8bit(n_req)); break;
  }
  return 0;
}

/**
 * TDS data Header class
 */

class TDS_HD{
  uint8_t data[28]; ///< header data
  int32_t day_cnt;  ///< Day counter starting from B.of File.
  // day_cnt is not use for VSSP32
 public:
  /// Constructor 
  TDS_HD():day_cnt(0){memset(data,0,4);}
//   const int Set_Data(unsigned char str[4], 
// 		     unsigned int pre_sod=0){
//     strncpy((char *)data, (char *)str, 4); 
//     if(pre_sod > Sod() && Sod()<120){ day_cnt += 1; return 1;} 
//     else return 0; 
//   }
  /// Set Header data
  const int Set_Data(unsigned char str[28], unsigned int pre_sod=0);
  /// Returns Second of the Day
  const unsigned int Sod(void) const;
  /// Returns Day of Year count 
  unsigned int Day_cnt(void){return(day_cnt);} 
  /// Returns Sampling Frequency (Hz)
  const uint32_t Sfrq(void)const; 
  /// Returns Number of channels
  const unsigned int Nch(void) const; 
  /// Returns Quantization bit
  const unsigned int Bit(void) const;
  /// Returns Sync-byte character data
  const unsigned char SyncB(void) const {return(data[3]);}
  /// Returns pointer of the header data 
  const unsigned char* c_data(void) const {return data;}
  /// Print data information 
  void  Print_inf(FILE *fp=stderr, const int level=0);
  const int Doy(void);  ///< Returns Day of year
  const int Year(void); ///< Returns Year
  ///  Returns Frame error flag
  const int FlagFrameErr(void) const {
    return ((data[5] & 0x80) >> 7) == 0 ? 0 : 1;}
  /// Returns Aux frame size 
  const uint8_t AuxSize(void)const {return data[6];}
  /// Returns Major Version Number 
  const uint8_t MajVer(void) const {return ((data[7] & 0xf0) >> 4);}
  /// Returns Miner Version Number 
  const uint8_t MinVer(void)const {return (data[7] & 0x0f);}
  /// Returns Filter Frequency  (MHz)
  const uint8_t FilterFreq(void)const {return data[9];}
  const uint8_t FormatNum(void)const {return data[8];} ///< Format Number
  const string  StnID(void); ///< Station ID 
  const string  Stname(void); ///< Returns Station name 
  const string  Hostname(void); ///< Returns hostname in string
  const time_t  Epoch();  ///< Returns Epoch of the data
};

inline const time_t  TDS_HD::Epoch(){
  int yy =Year();
  int doy = Doy();
  if(yy < 0 || doy < 0){
    cerr << "Error to get TDS_HD::Year() or TDS_HD::Doy()\n";
    return 0;
  }
  uint32_t sod = Sod();
  struct tm tm1;
  tm1.tm_sec=0; tm1.tm_min=0;  tm1.tm_hour=0;
  tm1.tm_mday = 1;  tm1.tm_mon = 0;  tm1.tm_year = yy-1900;
  time_t t1 = mktime(&tm1) + (doy-1)*86400 + sod;
  return t1;
}

inline const string TDS_HD::Stname(void){
  if(FormatNum() != 1){
    fprintf(stderr, "Error in TDS_HD::Stname(): FormatNumber(%d) is not 1\n",
	    data[8]);
    return "";
  }
 char buf[9]; 
 memcpy(buf,data+12,8); 
 buf[8] = '\0';
 string ret = buf;
 return ret;
}

inline const string TDS_HD::Hostname(void){
  if(FormatNum() != 1){
    cerr << "Error in TDS_HD::Hostname(): FormatNumber is not 1\n";
    return "";
  }
  char buf[9]; 
  memcpy(buf,data+20,8); buf[8] = '\0';
  string ret = buf;
  return ret;
}

inline const string TDS_HD::StnID(void){
  if(FormatNum() != 1){
    cerr << "Error in TDS_HD::StnID(): FormatNumber is not 1\n";
    return "";
  }
  char buf[3] = {data[10],data[11],'\0'};
  string ret = buf;
  return ret;
}

inline const int TDS_HD::Doy(void){
  if(data[3] != 0x8c){
    cerr << "Error in TDS_HD::Doy(): Requested Doy() but this is not VSSP32 data" << endl;
    return -1;
  }
  int doy=0;
  doy |= (uint32_t)(data[5] & 0x01) << 8;
  doy |= (uint32_t)(data[4] & 0xff);
  return doy;
}

inline const int TDS_HD::Year(void){
   if(data[3] != 0x8c){
    cerr << "Error in TDS_HD::Year(): Requested Year() but this is not VSSP32 data" << endl;
    return -1;
  }
  int yy=0;
  yy |= (uint32_t)(data[5] & 0x7e) >>1;
  yy += yy < 80 ? 2000 : 1900;
  return yy;
}

// inline void TDS_HD::Print_inf(FILE *fp){
// //   if(data[3] == 0x8b || data[3] == 0x8c )
// //     fwrite("HD_Symbol OK\n",13,1,fp);
// //   else
//   switch(data[3]){
//   case 0x8b: 
//     //    fwrite("   HD_Type=VSSP\n",16,1,fp);
//     fprintf(fp,"   HD_Type=VSSP\n");
//     break;
//   case 0x8c:
//     //    fwrite("   HD_Type=VSSP32\n",18,1,fp);
//     fprintf(fp,"   HD_Type=VSSP32\n");
//     break;
//   default:
//     //    fwrite("   HD_Symbol NG\n",16,1,fp);
//     fprintf(fp,"   HD_Symbol NG\n");
//     break;
//   }
// }

inline const int TDS_HD::Set_Data(unsigned char str[28], 
				  unsigned int pre_sod){
  switch(str[3]){
  case 0x8b:   // VSSP headder 
    //    strncpy((char *)data, (char *)str, 4); 
    memcpy((char *)data, (char *)str, 4); 
    if(pre_sod > Sod() && Sod()<120){ day_cnt += 1; return 1;} 
    break;
  case 0x8c: // VSSP32 headder   
    //    strncpy((char *)data, (char *)str, 28); 
    memcpy((char *)data, (char *)str, 28); 
    break;
  default:
    cerr << "Error in TDS_HD::Set_Data(): Sync-pattern2 does not maches\n";
    return -1;
    break;
  }
  return 0;
}

inline const unsigned int TDS_HD::Bit(void) const {
  unsigned int bit_flag = ((data[2] & 0xc0) >> 6);
  unsigned int nbit=1;
  switch (bit_flag){
  case 0:
    nbit = 1;
    break;
  case 1:
    nbit = 2;
    break;
  case 2:
    nbit = 4;
    break;
  case 3:
    nbit = 8;
    break;
  default:
    nbit=0;
    break;
  }
  return(nbit);
}
inline const unsigned int TDS_HD::Nch(void) const{
  int iflg = (data[2] & 0x02) >> 1;
  if(iflg ==0) return(1);
  else         return(4);
}
inline const uint32_t TDS_HD::Sfrq(void) const{
  unsigned int ifrq = (data[2] & 0x3c) >> 2;
  uint32_t sfrq=40000;
  
  switch( ifrq ){
  case 0:
    sfrq = 40000;
    break;
  case 1:
    sfrq = 100000;
    break;
  case 2:
    sfrq = 200000;
    break;
  case 3:
    sfrq = 500000;
    break;
  case 4:
    sfrq = 1000000;
    break;
  case 5:
    sfrq = 2000000;
    break;
  case 6:
    sfrq = 4000000;
    break;
  case 7:
    sfrq = 8000000;
    break;
  case 8:
    sfrq = 16000000;
    break;
  case 9:
    sfrq = 32000000;
    break;
  default:
    sfrq = 0;
    break;
  }
  return (sfrq);
  
}

inline const unsigned int TDS_HD::Sod(void) const{
  uint32_t sod=0;
  sod |= (data[2] & 0x001 ) << 16 ;
  sod |= (uint32_t)(data[1] & 0xff) << 8;
  sod |= (uint32_t)(data[0] & 0xff) ;
  
  //  sod |= (data[2] & 0x001 ) << 16 ;
  return (sod);
}

#endif

/*
 * $Log: tds_samplr.hpp
 *  Author: $Author: sekido $
 *  Date: $Date: 2010/01/02 03:40:37 $
 *  $Header: /vsspfs1/work3/cvs-archive/VDIF/tds_cls.hpp,v 1.2 2010/01/02 03:40:37 sekido Exp $
 *  Revision: $Revision: 1.2 $
 *  Source: $Source: /vsspfs1/work3/cvs-archive/VDIF/tds_cls.hpp,v $
 */
